/*
    Beschleunigungsmessung mit Arduino UNO und Adafruit Datalogging Shield

    Wasja Kleider
    Juli 2015
    Universität Heidelberg
    
    Speicherpfad auf SD-Karte: "A_LOG_XX.DAT" (XX=fortlaufende Nummer 00-99)

    Datenformat: "t (ms) \t t_RTC \t ACC_X (m/s^2) \t ACC_Y (m/s^2) \t ACC_Z (m/s^2)"
    Beispiel:    "2682  2015/7/9 13:52:49 1,31  0,92  -9,77" (Tab-separierte Spalten)
    
    Dieser Sketch verwendet eine veränderte Version der Library BMA020.h von Robin Gerhartz.
*/

// Einstellungen für Messungen NUR hier ändern:
#define measureInterval 20      // Zeitabstand der Einzelmessungen in Millisekunden, Standard=2000 (2s).
#define measureRange RANGE_8g   // Messbereich: RANGE_2g, RANGE_4g oder RANGE_8g (Standard).
#define baudRate 115200         // Baudrate für die Serielle Kommunikation, Standard = 115200.

// Für das Standardprogramm: Ab hier keine Änderungen mehr vornehmen!

// nötige Libraries laden
#include "Wire.h"   // I2C-Kommunikation (Beschleunigungssensor, RealTimeClock)
#include "BMA020.h" // Beschleunigungssensor
#include "SPI.h"    // SPI-Kommunikation für SD-Karte
#include "SD.h"     // SD-Karte
#include "RTClib.h" // RealTimeClock

// Variablen für SD-Kommunikation
const int CS = 10; // SPI-Chip Select
// weitere SPI-Pins: MOSI 11, MISO 12, CLK 13 -- auf Datalogging Shield vorgegeben

// Datei, in der die Daten gespeichert werden
File logFile;

// Pinbelegungen für das Bedienungsfeld
const int errorLED = 7;     // LED fuer Anzeige eines Fehlers bei Initialisierung oder Messung
const int measureLED = 6;   // LED fuer Anzeige einer Messwertaufnahme
const int stopButton = 5;   // Button für Messstopp
const int startButton = 4;  // Button für Messstart

// Variablen für die Messung
//const uint8_t measureRange = RANGE_8g; // Messbereich: RANGE_2g, RANGE_4g (Standard) oder RANGE_8g
unsigned long currentTimestamp = 0; // Zeitmarke fuer Messungen
bool measurementInProgress = false;
char filenamesave[] = "TESTTEST.DAT";

// Echtzeituhr
RTC_DS1307 RTC;

void setup() {
  led_both();
  Serial.begin(baudRate);
  Serial.println("Programmstart.");
  
  // Einrichten der LED und der Buttons
  pinMode(errorLED, OUTPUT);
  pinMode(measureLED, OUTPUT); 
  pinMode(startButton, INPUT_PULLUP); // Pullup-Widerstand aktiv, sodass LOW einem gedrueckten Knopf entspricht
  pinMode(stopButton, INPUT_PULLUP);  // s.o.
  
  // Starten des Beschleunigungssensors im gewählten Messbereich
  BMA::begin(measureRange, BANDWIDTH_50hz);

  // Kommunikation mit der SD-Karte aufbauen  
  pinMode(CS, OUTPUT); // Chip Select als Output
  if (!SD.begin(CS)) {
    error("Keine SD-Karte erkannt."); // Fehlermeldung anzeigen und Programm anhalten
  }
  Serial.println("SD-Karte initialisiert.");
  
  // Erstelle ein neues Logfile und öffne es.
  char filename[] = "A_LOG_00.DAT";
  for (uint8_t i = 0; i < 100; i++) {
    filename[6] = i/10 + '0';
    filename[7] = i%10 + '0';
    if (! SD.exists(filename)) {
      // Nur öffnen, falls die Datei noch nicht existiert.
      logFile = SD.open(filename, FILE_WRITE);
      
      // Speichere auch den endgültigen Dateiname, damit Daten gespeichert werden können.
      for (int j = 0; j<14; j++) {
        filenamesave[j] = filename[j];
      }
      break;  // Dann Schleife beenden.
    }
  }
  if(!logFile) {
    error("Log-Datei konnte nicht angelegt werden. (bis 99?)"); // Evtl. alte Logfiles löschen!
  }
  else {
    Serial.print("Log-Datei: ");
    Serial.println(filenamesave);
  }

  // Starte die Kommunikation mit der Echtzeituhr
  Wire.begin();  
  if (!RTC.begin()) {
    logFile.println("RTC-Fehler.");
    error("RTC-Fehler.");
  }
  
  // Schreibe eine Kopfzeile in die Datei
  logFile.println("t(ms)\t t_RTC\t ACC_X(m/s^2)\t ACC_Y(m/s^2)\t ACC_Z(m/s^2)");
  
  // Schließe das Logfile, damit die bisherigen Daten gespeichert werden.
  logFile.close();
  
  // Signalisiere Ende der Initialisierung durch Ausschalten beider LED
  delay(1000);
  led_none();
}


// Hier folgt das eigentliche Messprogramm.
void loop() {
  
  // Falls keine Messung läuft, warte bis der Start-Button gedrückt wird.
  if(!measurementInProgress) {
    while(digitalRead(startButton)) {
        // Start-Button wurde nicht gedrückt, also warte.
    }
    measurementInProgress = true;
    logFile = SD.open(filenamesave, FILE_WRITE);
    if(!logFile) {
      error("Speicherpfad nicht verfügbar. (MESS-START)");
    }
    Serial.println("Messung gestartet.");
    led_measure();
  }
  // Falls eine Messung läuft und der Stop-Button gedrückt wird, halte die Messung an.
  else if(measurementInProgress) {
    if (!digitalRead(stopButton)) {
      // Button ist auf LOW, also gedrueckt
      measurementInProgress = false;
      logFile.close(); // Daten werden in das Logfile geschrieben.
      Serial.println("Messung beendet.");
      led_both();
      delay(500);
      led_none();
      delay(500);
      led_both();
      delay(500);
      led_none();
      return; // Beende die Schleife vorzeitig, sodass auf ein erneutes Startsignal gewartet wird.
    }
  }
  
  // Nimm die gewünschten Zeiten und Messdaten auf.
  currentTimestamp = millis();
  DateTime now = RTC.now();
  BMA::read_all_axes();
  
  // Schreibe alle Daten in das Logfile.
  logFile.print(currentTimestamp);
  logFile.print("\t");
  logFile.print(makeTimestamp(now));
  logFile.print("\t");
  logFile.println(getAccelerationData());
//  Serial.print(currentTimestamp);
//  Serial.print("\t");
//  Serial.print(makeTimestamp(now));
//  Serial.print("\t");
//  Serial.println(getAccelerationData());

  // Warte, bis der nächste Messwert aufgenommen werden soll.
  while(millis() < (currentTimestamp+measureInterval)) 
  {
    // Falls während der Wartezeit der Stop-Button gedrückt wird, nicht länger warten!
    if(!digitalRead(stopButton)) return;
  }
}







// Schalte beide LED ein.
void led_both() {
  digitalWrite(errorLED, HIGH); // Fehler-LED einschalten
  digitalWrite(measureLED, HIGH); // Messungs-LED einschalten
}

// Schalte nur die Error-LED ein.
void led_error() {
  digitalWrite(errorLED, HIGH); // Fehler-LED einschalten
  digitalWrite(measureLED, LOW); // Messungs-LED ausschalten
}

// Schalte nur die Mess-LED ein.
void led_measure() {
  digitalWrite(errorLED, LOW); // Fehler-LED ausschalten
  digitalWrite(measureLED, HIGH); // Messungs-LED einschalten  
}

// Schalte beide LED aus.
void led_none() {
  digitalWrite(errorLED, LOW); // Fehler-LED ausschalten
  digitalWrite(measureLED, LOW); // Messungs-LED ausschalten
}



// Überführe die aktuelle Zeit in einen lesbaren String des Formats "YYYY/MM/DD HH:MM:SS"
String makeTimestamp(DateTime time) {
  String returnString = "";
  
  returnString += (String) time.year();
  returnString += "/";
  returnString += (String) time.month();
  returnString += "/";
  returnString += (String) time.day();
  returnString += " ";
  returnString += (String) time.hour();
  returnString += ":";
  returnString += (String) time.minute();
  returnString += ":";
  returnString += (String) time.second();
  
  return returnString;
}


// Liefere die Beschleunigungsdaten der drei Achsen skaliert auf den Messbereich
String getAccelerationData() {
  String returnString = "";
  
  switch (measureRange) {
    case RANGE_2g: 
      returnString += translateToG_2G(BMA::X);
      returnString += "\t";
      returnString += translateToG_2G(BMA::Y);
      returnString += "\t";
      returnString += translateToG_2G(BMA::Z);
    break;
    
    case RANGE_4g:
      returnString += translateToG_4G(BMA::X);
      returnString += "\t";
      returnString += translateToG_4G(BMA::Y);
      returnString += "\t";
      returnString += translateToG_4G(BMA::Z);
    break;
    
    case RANGE_8g:
      returnString += translateToG_8G(BMA::X);
      returnString += "\t";
      returnString += translateToG_8G(BMA::Y);
      returnString += "\t";
      returnString += translateToG_8G(BMA::Z);
    break;
    
    default:
      // keiner der oben gewählten Bereiche -- Fehler!
      error("Messbereich BMA020 fehlerhaft eingestellt.");
    break;
  }
  
  returnString.replace(".",","); // Ersetzen der amerikanischen Dezimalstelle (Punkt) durch deutsche Dezimalstelle (Komma).
  return returnString;
}

// Uebersetze das Messergebnis im Messbereich 2g in g (Einheit m/s^2) 
float translateToG_2G(int value) {
  float val = map(value, -255, 255, -9810, 9810);
  return val/1000;
}

// Uebersetze das Messergebnis im Messbereich 4g in g (Einheit m/s^2) 
float translateToG_4G(int value) {
  float val = map(value, -255, 255, -19620, 19620);
  return val/1000;
}

// Uebersetze das Messergebnis im Messbereich 8g in g (Einheit m/s^2) 
float translateToG_8G(int value) {
  float val = map(value, -255, 255, -39240, 39240);
  return val/1000;
}


// Falls ein Fehler im Programmablauf auftritt, sende eine Fehlernachrit an die
// serielle Schnittstelle, schalte die Fehler-LED ein und halte das Programm an.
void error(char *str)
{
  Serial.print("Fehler: ");
  Serial.println(str);
  led_error();
  while(1);
}
